/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is SmoothWheel.
 *
 * The Initial Developer of the Original Code is Avi Halachmi.
 * Portions created by the Initial Developer are Copyright (C) 2003 Avi Halachmi.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*********************************************************
SmoothWheel by Avi Halachmi (avihpit (*at*) yahoo (*dot*) com)
http://smoothwheel.mozdev.org
Changelog, Info and PR statement is at the bottom of this file.
started  (Avi Halachmi): 22-Mar-2003 (v0.1)
modified (Avi Halachmi): 15-Apr-2003 (v0.2):
 - Improved Algorithm, Improved Timing, much improved target detection,
   Modifier keys, Better configuration and presets, Optimizations.
*********************************************************/


//**************** CONFIGURATION SECTION ****************

//Global enable/disable for this extension. default is true.
// setting it to false will disable this extension completely,
// overriding all of the following configurations values.
const EnableThisExtension=true;

// select one of several preset settings: (default is 3)
// 0 : don't use preset. use actual variables values.
// 1 : turbo:  for those who like quickies
//          (duration=150, non-linear, smooth transitions, relative step)
// 2 : fast: slower than turbo, faster than the default
//          (duration=300, non-linear, smooth transitions, relative step)
// 3 : medium: default settings
//          (duration=500, non-linear, smooth transitions, relative step)
// 4 : nice and easy: same as medium, but slower and smoother
//          (duration=1000, non-linear, smooth transitions, relative step)
// 5 : same as version 0.1 default (duration=250, linear, no smooth transition, no relative step)
// 6 : same as version 0.1 with isLinear=false, duration=150 (no smooth transition, no relative step)
const PresetSettings=3;

//IMPORTANT NOTE ABOUT MODIFIER KEYS:
//by default, ctrl key is assigned to enlarge/reduce the text size.
//the alt key is assigned to browse history. shift key is not assigned.
//in order for a modifier key to work with this extension, it has to be UNassigned first.
//you can UNassign a modifier key for the mouse wheel as follows:
//mozilla: edit->preferences->advanced->mouse Wheel -> set it to scroll normally.
//phoenix/firebird: about:config in the locationbar bar. then change mousewheel.with<xxx>key.action property to 0

//Select a modifier key for fine small scroll steps. default is 3 (SHIFT key).
//possible values: 0=Disable, 1=ALT, 2=CTRL, 3=SHIFT
const ModifierSmallStep=3;

//Select a modifier key for full page scroll (like PgUp/PgDn). default is 1 (ALT key).
//possible values: 0=Disable, 1=ALT, 2=CTRL, 3=SHIFT
const ModifierFullPageStep=1;

//max milisec to complete scroll after last scroll event
//reasonable values are (fast)100/250/500/1000(slow). default is 500
// -> this value is overriden if PresetSettings<>0
var   ScrollDuration=500;

//if enabled, set step size relative to the scrolled area height by a constant factor.
//thus, if enabled and the window/frame is short, so will be the scroll steps.
//when enabled, it overrides gecko internal fixed pixels/line constant. default is true.
// -> this value is overriden if PresetSettings<>0
var   EnableRelativeStepSize=true;

//Limits the frames/sec of the smooth scroll operation.
//for best visual performance, set this value to your monitor refresh rate. default is 100.
//reasonable values are 20 to 120
//note that setting it to high values will make no difference if your cpu isn't up for the task.
//tips: for faster Gecko rendering, set Desktop color to 16bpp, and turn off font-anti-aliasing
const ScrollMaxFPS=100;

//********* END OF COMMON CONFIGURATION OPTIONS SECTION ***********


//Whether or not to enable WHOLE PAGE scroll when mouse wheel is over form textarea. default is false.
// if set to false, mouse wheel events will NOT be intercepted by this extension while
//   mouse is over a form textarea. instead, the DEFAULT scroll will be issued.
// if set to true, the WHOLE PAGE will scroll smoothly when mouse is over a form textarea.
const EnableWholePageScrollInForms=false;

//selects either linear (=0) or various non-linear(=1/2/3) scroll profiles. default is 1.
//possible values:
// 0 : Linear
// 1 : 2nd power (non-linear)
// 2 : 3rd power (non-linear)
// 3 : cosine    (non-linear)
// -> this value is overriden if PresetSettings<>0
// -> changing the default value will prevent code optimizations.
var   SmoothScrollProfile=1;

//tries to make a smooth transition if new scroll event occures while still scrolling. default is true.
// -> this value is overriden if PresetSettings<>0
// -> changing the default value will prevent code optimizations.
var   EnableSmoothEventTransition=true;

//factor to use when EnableRelativeStepSize is true. default is 6.7
const RelativeStepFactor=6.7;

//factor for how much shorter the scroll steps are if ModifierSmallStep is used. default is 10.
const ShortStepFactor=10;

//#pixels per scroll line. can i read this value from anywhere???
//default value is 20 (empirical estimation of internal Gecko value)
const ScrollPx=20;

//percentage of when to finish transitional profile (from previous velocity to absolute scroll profile)
//default is 1.
// -> this value is overriden if PresetSettings<>0
// -> changing the default value will prevent code optimizations.
var   TransEnd=1;

//*************** END OF CONFIGURATION SECTION ********************

const LoopInterval=1000/ScrollMaxFPS; //fps converted to ms between scroll events

//Global variables (used by more than one function).
var scrollLeft=0; //in pixels
var scrollLeftLastEvent=0; //in pixels
var scrollEventTimeStamp;
var scrollDeadline;// timestamp
var clientFrame=null;  //holds the object to scroll
var intervalID=null;   //holds the interval timer handle
var velocityLastEvent=0;
var velocityCurrent=0;
var scrollCurrentStep;  //how many pixels to scroll for the current scroll event.
var optimizedProfileCodeExists; //whether optimized code exist according to various variables


//return a string of all parentNode elements (recursively) for the input node
function parentTree(node){
  if (!node)
     return "[object null(generatedString)]";

  return ""+parentTree(node.parentNode)+ "," + node;
}

//checks whether the specified modifier key was held when the event was intercepted.
function checkModifier(e, modifier){
    switch (modifier) {
       case 1: return e.altKey;
       case 2: return e.ctrlKey;
       case 3: return e.shiftKey;
    }
    return false;
}

//called when wheel event captured, while NOT scrolling.
//determines whether to leave it alone, or to use it (and initilize the vars)
function acquireEvent(e){
    var parents= parentTree(e.target);
    var tmpClientFrame= e.target.ownerDocument.defaultView;

    //1st, check if we should intercept this event at all, or let the default handler do it's stuff.
    if (
        tmpClientFrame != "[object Window]" ||
        parents.indexOf("HTMLTextAreaElement")>=0  && (!EnableWholePageScrollInForms) ||
        parents.indexOf("XULElement")>=0 ||
        parents.indexOf("HTMLEmbedElement")>=0 ||
        //parents.indexOf("HTMLImageElement")>=0 ||
        parents.indexOf("HTMLOptionElement")>=0
       )
          return false; //default handler will take it from here.

    //implicit else: ok. intercepting event. initialize variables, step size etc.
    intervalID= window.setInterval("smoothWheelLoop()",LoopInterval);
    clientFrame= tmpClientFrame;

    var viewHeight= clientFrame.innerHeight;
    if (checkModifier(e, ModifierFullPageStep))
       scrollCurrentStep=viewHeight;
    else
       scrollCurrentStep=(EnableRelativeStepSize)? viewHeight/RelativeStepFactor: ScrollPx*(Math.abs(e.detail));

    if (checkModifier(e, ModifierSmallStep))
       scrollCurrentStep= scrollCurrentStep/ShortStepFactor;

    velocityCurrent=0;
    return true;
}


//SmoothWheel wheel event handler, registered for handling DOMMouseScroll events
function smoothWheelHandler(e){
      if (!intervalID) //if not currently scrolling
         if (!EnableThisExtension || !acquireEvent(e)) //if event should not be intercepted
            return;  //let default handler do it's stuff
         //implicit else: event acquired and vars initialized

      e.preventDefault();
      e.stopPropagation();

      scrollEventTimeStamp= (new Date()).getTime();
      scrollDeadline= scrollEventTimeStamp + ScrollDuration;

      scrollLeft += ((e.detail>0)?1:-1)*scrollCurrentStep;
      scrollLeftLastEvent= scrollLeft;
      velocityLastEvent= velocityCurrent;
}


//loop for smooth scrolling the page
function smoothWheelLoop(){
      var toScroll, currentTime, currentPercentage, timeLeft;
      currentTime= (new Date()).getTime();
      timeLeft= scrollDeadline - currentTime;

      if (timeLeft<=0)  //we're past the deadline. scroll what's left.
          toScroll=scrollLeft;
      else{ //calculate next scroll destination according to scroll profile
          currentPercentage= 1 - timeLeft/ScrollDuration;
          toScroll= Math.round(scrollLeftLastEvent*(scrollProfile(currentPercentage) - 1) + scrollLeft);
          if ((scrollLeft-toScroll)*scrollLeft < 0) //if going past destination point
              toScroll=scrollLeft;
      }

      clientFrame.scrollBy(0, toScroll);
      scrollLeft -= toScroll;
      if (scrollLeft==0){//no more scroll needed
         window.clearInterval(intervalID);
	 intervalID=null;
         velocityCurrent=0;
         return;
      }
}


//declerations for optimized code only.
var p,p1,p2,p21,p2v;

//transform percentage of time to percentage of location according to a scroll profile
//it might consider an implicit parameter of the speed of the scroll at the time the event was intercepted.
function scrollProfile(percentage){
        if (optimizedProfileCodeExists){
             p=percentage; p1=1-p; p2=p1*p1; p21=1-p2; p2v=p2*velocityLastEvent;
             velocityCurrent= p2v + p21*2*p1;
             return p2v*p + p21*p21; //location
        }
        // the above optimized code (hardcoded values, inlined functioned) is used if:
        //    SmoothScrollProfile=1 (->profilePower2) and EnableSmoothEventTransition=true and TransEnd=1
        // the optimized code does the stuff instead of running fully (sometimes multiple times) through the next functions:
        //    scrollProfile(p), transition(p,x,y),
        //    profilePower2(p), profilePower2Velocity(p),
        //    profileConstVelocity(p), profileConstVelocityVelocity(p)
        // the best optimization though would be to get 100-200 precalculated results into a table, and merge the value with velocityLastEvent in runtime;)
        //optimized transition(p,x,y)=p2*x+p21*y //p2 and p21 are already functions of p

        //implicit else: non-optimized path, consider all settings explicitly.
        var location, velocity;
        switch (SmoothScrollProfile){
               case 0:  location= profileLinear(percentage); break;
               case 1:  location= profilePower2(percentage); break;
               case 2:  location= profilePower3(percentage); break;
               case 3:  location= profileCosine(percentage); break;
               default: location= profileLinear(percentage); break;
        }
        if (EnableSmoothEventTransition){
           switch (SmoothScrollProfile){
               case 0:  velocity= profileLinearVelocity(percentage); break;
               case 1:  velocity= profilePower2Velocity(percentage); break;
               case 2:  velocity= profilePower3Velocity(percentage); break;
               case 3:  velocity= profileCosineVelocity(percentage); break;
               default: velocity= profileLinearVelocity(percentage); break;
           }
           location= transition(percentage, profileConstVelocity(percentage), location);
           velocityCurrent= transition(percentage, profileConstVelocityVelocity(percentage), velocity);
        }
        return location;
}


//returns a weighted average of 2 profiles according (non-linear) to percentage.
function transition(percentage, srcProfile, destProfile){
         var srcPer= ((TransEnd-percentage)/TransEnd)*((TransEnd-percentage)/TransEnd);
         var destPer= 1-srcPer;
         return (percentage < TransEnd)?
                        srcPer*srcProfile + destPer*destProfile: destProfile;
}


//various scroll profiles and their velocities (derivatives) functions.
function profileLinear(percentage){return percentage;}
function profileLinearVelocity(percentage){return 1;}

function profilePower2(percentage){return 1-(1-percentage)*(1-percentage);}
function profilePower2Velocity(percentage){return 2*(1-percentage);}

function profilePower3(percentage){return 1-(1-percentage)*(1-percentage)*(1-percentage);}
function profilePower3Velocity(percentage){return 3*(1-percentage)*(1-percentage);}

function profileCosine(percentage){return .5-.5*Math.cos(Math.PI*percentage);}
function profileCosineVelocity(percentage){return .5*Math.sin(Math.PI*percentage)*Math.PI;}

//using implicit parameter velocityLastEvent.
//this profile does NOT achieve destination point on deadline.
//it represents a linear motion at the speed that was recorded when the event was intercepted.
//this profile is used for transition when a scroll event is intercepted while still scrolling.
function profileConstVelocity(percentage){return velocityLastEvent*percentage;}
function profileConstVelocityVelocity(percentage){return velocityLastEvent;}


//initialize variable according to the preset, and determine whether optimization is possible.
function initSmoothWheel(){
    //[ScrollDuration, EnableRelativeStepSize, SmoothScrollProfile, EnableSmoothEventTransition, TransEnd]
    var presets=[, [150,  true,  1, true,  1],
                   [300,  true,  1, true,  1],
                   [500,  true,  1, true,  1],
                   [1000, true,  1, true,  1],
                   [250,  false, 0, false, 1],
                   [150,  false, 2, false, 1]
                 ];

    if (PresetSettings >= 1 && PresetSettings <=6){
       ScrollDuration=              presets[PresetSettings][0];
       EnableRelativeStepSize=      presets[PresetSettings][1];
       SmoothScrollProfile=         presets[PresetSettings][2];
       EnableSmoothEventTransition= presets[PresetSettings][3];
       TransEnd=                    presets[PresetSettings][4];
    }

    optimizedProfileCodeExists= (SmoothScrollProfile==1) && (EnableSmoothEventTransition==true) &&  (TransEnd==1);
}

//runs once, when the file is loaded.
initSmoothWheel();

/***************************************************************************
Smooth Wheel extension for mozilla/phoenix/firebird
SmoothWheel home page: http://smoothwheel.mozdev.org
Download latest version: http://downloads.mozdev.org/smoothwheel/smoothwheel.xpi
Licence: MPL 1.1/GPL 2.0/LGPL 2.1
Copyright 2003 Avi Halachmi (avihpit (*at*) yahoo (*dot*) com)
Current version: v0.2 (15-Apr-2003)

About:
The smoothwheel extension will scroll the page smoothly with the mouse wheel.
It can enhance/replace the internal Gecko smoothscroll which was introduced on 2003-03-24.
It's also working with older versions of phoenix/mozilla on win32/linux.

Features:
- Very configurable, including 6 presets.
- Scroll steps relative to scrollable area height (configurable).
- Modifier keys for small/full page steps (SHIFT and ALT respectively by default).
- Advanced Optimized algorithm for a VERY smooth scroll in many scenarios.
- Timer based operation: Always 'finishes on time'.
- Automatically utilizes strong CPUs for smoother scroll.
- Can be disabled easily, even in mozilla.
- Lets the default mouse wheel handler do the job if smoothwheel can't handle it.

Todo:
- Make it work on XUL panes (phoenix/firebird view-source, mailAndNews, irc, etc).
- Possibly improve installation / chrome registration process.

/////////////////////////////////////////////////////////////////////////////
changelog:
/////////////////////////////////////////////////////////////////////////////
v0.2 (15-Apr-2003)

bugFix (v0.2):
- much better target detection code. will let default scroll handler do it's stuff if we can't handle properly.
- better timing of scroll loop
- fps limit (for fast CPUs)

new features (v0.2):
- modifier keys for slow/fullPage scroll
- improved algorithm: smooth transition if scroll event occures while still scrolling
- much more configuration options (including 6 presets)
- optimized code for the most common settings


status (v0.2):

scrolling SMOOTHLY on:
- mozilla: browser, view-source, irc (main window), mail-and-news (message windows)
- phoenix: browser

letting default scroll handler do it's stuff on:
- mozilla: browser forms, some mail-and-news panes, some irc panes,
           composer, all XUL panes, about:config, more..
- phoenix: browser forms, view-source, about:config, all XUL panes, more..

breaking wheel scroll functionality on (v0.2):
- couldn't find any so far.

/////////////////////////////////////////////////////////////////////////////
v0.1 (22-Mar-2003)

status (v0.1):

scrolling SMOOTHLY on:
- mozilla: browser, view-source, irc (main window), mail-and-news (message windows)
- phoenix: browser

letting default scroll handler do it's stuff on:
- mozilla: composer
- phoenix: view-source

breaking wheel scroll functionality on:
- mozilla: browser forms, some mail-and-news panes, some irc panes, all XUL panes, more...
- phoenix: browser forms, all XUL panes, more...
***************************************************************************
*/
